import numpy as np
cimport numpy as np
cimport cython


cpdef _omp_init():
    """
    Seems to be an issue with `-fopenmp` on macOS llvm clang which looks like this:
    ```
ImportError: dlopen(/mypylib/site-packages/kalmantv/eigen/kalmantv.cpython-38-darwin.so, 2): Symbol not found: ___kmpc_fork_call
  Referenced from: /mypylib/site-packages/kalmantv/eigen/kalmantv.cpython-38-darwin.so
  Expected in: flat namespace
 in /mypylib/site-packages/kalmantv/eigen/kalmantv.cpython-38-darwin.so
    ```
    Goes away if other cython code is loaded first, so that's what this function is for.
    """
    cdef int x = 0
    return x



from blas cimport *
import os
import numpy as np
cimport numpy as np


DTYPE = np.double

cpdef void state_sim(double[::1] x_state,
                     double[::1, :] llt_state,
                     const double[::1] mu_state,
                     const double[::1, :] var_state,
                     const double[::1] z_state):
    """
    Simulates from a normal distribution with mean `mu_state`, variance `var_state`,
    and randomness `z_state` drawn from :math:`N(0, 1)`.

    Args:
        x_state (ndarray(n_state)): Simulated state.
        llt_state (ndarray(n_state, n_state)): Temporary matrix to store cholesky factorization.
        mu_state (ndarray(n_state)): Mean vector.
        var_state (ndarray(n_state, n_state)): Variance matrix.
        z_state (ndarray(n_state)): Random vector simulated from :math:`N(0, 1)`.

    Returns:
        (tuple):
        - **x_state** (ndarray(n_state)): Simulated state.
        - **llt_state** (ndarray(n_state, n_state)): Temporary matrix to store cholesky factorization.

    """
    cdef char * trans = 'N'
    cdef char * uplo = 'L'
    cdef char * diag = 'N'
    cdef int x_alpha = 1
    chol_fact(llt_state, var_state)
    vec_copy(x_state, z_state)
    tri_vec_mult(x_state, uplo, trans, diag, llt_state)
    vec_add(x_state, x_alpha, mu_state)
    return

cdef class KalmanTV:
    r"""
    Create a Kalman Time-Varying object. The methods of the object can predict, update, sample and 
    smooth the mean and variance of the Kalman Filter. This method is useful if one wants to track 
    an object with streaming observations.

    The specific model we are using to approximate the solution :math:`x_n` is

    .. math::

        x_n = Q(x_{n-1} -\lambda) + \lambda + R_n^{1/2} \epsilon_n

        y_n = d + W x_n + \Sigma_n^{1/2} \eta_n

    where :math:`\epsilon_n` and :math:`\eta_n` are independent :math:`N(0,1)` distributions and
    :math:`y_n` denotes the model interrogation (observation) at time n.

    The variables of the model are defined below in the argument section. The methods of this class
    calculates :math:`\theta = (\mu, \Sigma)` for :math:`x_n` and the notation for
    the state at time n given observations from k is given by :math:`\theta_{n|K}`.

    Args:
        n_state (int): Number of state variables.
        n_meas (int): Number of measurement variables.
        mu_state_past (ndarray(n_state)): Mean estimate for state at time n-1 given observations from 
            times [0...n-1]; :math:`\mu_{n-1|n-1}`. 
        var_state_past (ndarray(n_state, n_state)): Covariance of estimate for state at time n-1 given 
            observations from times [0...n-1]; :math:`\Sigma_{n-1|n-1}`.
        mu_state_pred (ndarray(n_state)): Mean estimate for state at time n given observations from 
            times [0...n-1]; denoted by :math:`\mu_{n|n-1}`. 
        var_state_pred (ndarray(n_state, n_state)): Covariance of estimate for state at time n given 
            observations from times [0...n-1]; denoted by :math:`\Sigma_{n|n-1}`.
        mu_state_filt (ndarray(n_state)): Mean estimate for state at time n given observations from 
            times [0...n]; denoted by :math:`\mu_{n|n}`. 
        var_state_filt (ndarray(n_state, n_state)): Covariance of estimate for state at time n given 
            observations from times [0...n]; denoted by :math:`\Sigma_{n|n}`.
        mu_state_next (ndarray(n_state)): Mean estimate for state at time n+1 given observations from 
            times [0...N]; denoted by :math:`\mu_{n+1|N}`. 
        var_state_next (ndarray(n_state, n_state)): Covariance of estimate for state at time n+1 given 
            observations from times [0...N]; denoted by :math:`\Sigma_{n+1|N}`.
        x_state_smooths (ndarray(n_state)): Sample solution at time n given observations from times [0...N];
            denoted by :math:`X_{n|N}`
        mu_state_smooth (ndarray(n_state)): Mean estimate for state at time n given observations from 
            times [0...N]; denoted by :math:`\mu_{n|N}`.
        var_state_smooth (ndarray(n_state, n_state)): Covariance of estimate for state at time n given 
            observations from times [0...N]; denoted by :math:`\Sigma_{n|N}`.
        x_state (ndarray(n_state)): Simulated state vector; :math:`x_n`.
        mu_state (ndarray(n_state)): Transition offsets defining the solution prior; denoted by :math:`\lambda`.
        wgt_state (ndarray(n_state, n_state)): Transition matrix defining the solution prior; denoted by :math:`Q`.
        var_state (ndarray(n_state, n_state)): Variance matrix defining the solution prior; denoted by :math:`R`.
        x_meas (ndarray(n_meas)): Interrogated measure vector from `x_state`; :math:`y_n`.
        mu_meas (ndarray(n_meas)): Transition offsets defining the measure prior; denoted by :math:`d`.
        wgt_meas (ndarray(n_meas, n_state)): Transition matrix defining the measure prior; denoted by :math:`W`.
        var_meas (ndarray(n_meas, n_meas)): Variance matrix defining the measure prior; denoted by :math:`\Sigma_n`.
        z_state (ndarray(n_state)): Random vector simulated from :math:`N(0, 1)`.
        mu_fore (ndarray(n_meas)): Mean estimate for measurement at n given observations from [0...n-1]
        var_fore (ndarray(n_meas, n_meas)): Covariance of estimate for state at time n given 
            observations from times [0...n-1]
    """

    def __init__(self, int n_meas, int n_state):
        self.n_meas = n_meas
        self.n_state = n_state
        self.tmu_state = np.empty(n_state, dtype=DTYPE)
        self.tmu_state2 = np.empty(n_state, dtype=DTYPE)
        self.tvar_state = np.empty((n_state, n_state),
                                   dtype=DTYPE, order='F')
        self.tvar_state2 = np.empty((n_state, n_state),
                                    dtype=DTYPE, order='F')
        self.tvar_state3 = np.empty((n_state, n_state),
                                    dtype=DTYPE, order='F')
        self.tmu_meas = np.empty(n_meas, dtype=DTYPE)
        self.tvar_meas = np.empty((n_meas, n_meas),
                                  dtype=DTYPE, order='F')
        self.twgt_meas = np.empty((n_meas, n_state),
                                  dtype=DTYPE, order='F')
        self.twgt_meas2 = np.empty((n_meas, n_state),
                                   dtype=DTYPE, order='F')
        self.llt_meas = np.empty((n_meas, n_meas),
                                 dtype=DTYPE, order='F')
        self.llt_state = np.empty((n_state, n_state),
                                  dtype=DTYPE, order='F')

    cpdef void predict(self,
                       double[::1] mu_state_pred,
                       double[::1, :] var_state_pred,
                       const double[::1] mu_state_past,
                       const double[::1, :] var_state_past,
                       const double[::1] mu_state,
                       const double[::1, :] wgt_state,
                       const double[::1, :] var_state):
        r"""
        Perform one prediction step of the Kalman filter.
        Calculates :math:`\theta_{n|n-1}` from :math:`\theta_{n-1|n-1}`.
        """
        cdef char * wgt_trans = 'N'
        cdef char * var_trans = 'N'
        cdef char * wgt_trans2 = 'T'
        cdef int mu_alpha = 1, mu_beta = 1, var_alpha = 1, var_beta = 1
        vec_copy(mu_state_pred, mu_state)
        mat_vec_mult(mu_state_pred, wgt_trans, mu_alpha,
                     mu_beta, wgt_state, mu_state_past)
        mat_copy(var_state_pred, var_state)
        mat_triple_mult(var_state_pred, self.tvar_state, wgt_trans, var_trans, wgt_trans2,
                        var_alpha, var_beta, wgt_state, var_state_past, wgt_state)

        return

    cpdef void update(self,
                      double[::1] mu_state_filt,
                      double[::1, :] var_state_filt,
                      const double[::1] mu_state_pred,
                      const double[::1, :] var_state_pred,
                      const double[::1] x_meas,
                      const double[::1] mu_meas,
                      const double[::1, :] wgt_meas,
                      const double[::1, :] var_meas):
        r"""
        Perform one update step of the Kalman filter.
        Calculates :math:`\theta_{n|n}` from :math:`\theta_{n|n-1}`.
        """
        cdef char * wgt_trans = 'N'
        cdef char * wgt_trans2 = 'T'
        cdef char * var_trans = 'N'
        cdef int tmu_alpha = -1, tmu_beta = -1, tvar_alpha = 1, tvar_beta = 1
        cdef int tmu_alpha2 = 1
        cdef int mu_alpha = 1, mu_beta = 1, var_alpha = -1, var_beta = 1
        vec_copy(self.tmu_meas, mu_meas)
        mat_vec_mult(self.tmu_meas, wgt_trans, tmu_alpha,
                     tmu_beta, wgt_meas, mu_state_pred)
        mat_copy(self.tvar_meas, var_meas)
        mat_triple_mult(self.tvar_meas, self.twgt_meas, wgt_trans, var_trans, wgt_trans2,
                        tvar_alpha, tvar_beta, wgt_meas, var_state_pred, wgt_meas)
        solveV(self.llt_meas, self.twgt_meas2, self.tvar_meas, self.twgt_meas)
        vec_add(self.tmu_meas, tmu_alpha2, x_meas)
        vec_copy(mu_state_filt, mu_state_pred)
        mat_vec_mult(mu_state_filt, wgt_trans2, mu_alpha,
                     mu_beta, self.twgt_meas2, self.tmu_meas)
        mat_copy(var_state_filt, var_state_pred)
        mat_mult(var_state_filt, wgt_trans2, wgt_trans, var_alpha, var_beta,
                 self.twgt_meas2, self.twgt_meas)
        return

    cpdef void filter(self,
                      double[::1] mu_state_pred,
                      double[::1, :] var_state_pred,
                      double[::1] mu_state_filt,
                      double[::1, :] var_state_filt,
                      const double[::1] mu_state_past,
                      const double[::1, :] var_state_past,
                      const double[::1] mu_state,
                      const double[::1, :] wgt_state,
                      const double[::1, :] var_state,
                      const double[::1] x_meas,
                      const double[::1] mu_meas,
                      const double[::1, :] wgt_meas,
                      const double[::1, :] var_meas):
        r"""
        Perform one step of the Kalman filter.
        Combines :func:`KalmanTV.predict` and :func:`KalmanTV.update` steps to get :math:`\theta_{n|n}` from :math:`\theta_{n-1|n-1}`.
        """
        self.predict(mu_state_pred, var_state_pred,
                     mu_state_past, var_state_past,
                     mu_state, wgt_state, var_state)
        self.update(mu_state_filt, var_state_filt,
                    mu_state_pred, var_state_pred,
                    x_meas, mu_meas, wgt_meas, var_meas)
        return

    cpdef void smooth_mv(self,
                         double[::1] mu_state_smooth,
                         double[::1, :] var_state_smooth,
                         const double[::1] mu_state_next,
                         const double[::1, :] var_state_next,
                         const double[::1] mu_state_filt,
                         const double[::1, :] var_state_filt,
                         const double[::1] mu_state_pred,
                         const double[::1, :] var_state_pred,
                         const double[::1, :] wgt_state):
        r"""
        Perform one step of the Kalman mean/variance smoother.
        Calculates :math:`\theta_{n|N}` from :math:`\theta_{n+1|N}`, :math:`\theta_{n|n}`, and :math:`\theta_{n+1|n}`.
        """
        cdef char * wgt_trans = 'N'
        cdef char * var_trans = 'T'
        cdef char * var_trans2 = 'N'
        cdef int tvar_alpha = 1, tvar_beta = 0, tmu_alpha = -1, tvar_alpha2 = -1, tvar_beta2 = 1
        cdef int mu_alpha = 1, mu_beta = 1, var_alpha = 1, var_beta = 1
        mat_mult(self.tvar_state, wgt_trans, var_trans, tvar_alpha, tvar_beta, wgt_state,
                 var_state_filt)
        solveV(self.llt_state, self.tvar_state,
               var_state_pred, self.tvar_state)
        vec_copy(self.tmu_state, mu_state_next)
        vec_add(self.tmu_state, tmu_alpha, mu_state_pred)
        mat_copy(self.tvar_state2, var_state_next)
        mat_add(self.tvar_state2, tvar_alpha2, tvar_beta2, var_state_pred)
        mat_mult(self.tvar_state3, var_trans, var_trans2, tvar_alpha, tvar_beta, self.tvar_state,
                 self.tvar_state2)
        vec_copy(mu_state_smooth, mu_state_filt)
        mat_vec_mult(mu_state_smooth, var_trans, mu_alpha,
                     mu_beta, self.tvar_state, self.tmu_state)
        mat_copy(var_state_smooth, var_state_filt)
        mat_mult(var_state_smooth, var_trans2, var_trans2, var_alpha, var_beta, self.tvar_state3,
                 self.tvar_state)
        return

    cpdef void smooth_sim(self,
                          double[::1] x_state_smooth,
                          const double[::1] x_state_next,
                          const double[::1] mu_state_filt,
                          const double[::1, :] var_state_filt,
                          const double[::1] mu_state_pred,
                          const double[::1, :] var_state_pred,
                          const double[::1, :] wgt_state,
                          const double[::1] z_state):
        r"""
        Perform one step of the Kalman sampling smoother.
        Calculates a draw :math:`x_{n|N}` from :math:`x_{n+1|N}`, :math:`\theta_{n|n}`, and :math:`\theta_{n+1|n}`.
        """
        cdef char * wgt_trans = 'N'
        cdef char * var_trans = 'T'
        cdef char * var_trans2 = 'N'
        cdef int tvar_alpha = 1, tvar_beta = 0, tmu_alpha = -1
        cdef int tmu_alpha2 = 1, tmu_beta2 = 1, tvar_alpha2 = -1, tvar_beta2 = 1
        mat_mult(self.tvar_state, wgt_trans, var_trans, tvar_alpha, tvar_beta, wgt_state,
                 var_state_filt)
        solveV(self.llt_state, self.tvar_state2,
               var_state_pred, self.tvar_state)
        vec_copy(self.tmu_state, x_state_next)
        vec_add(self.tmu_state, tmu_alpha, mu_state_pred)
        vec_copy(self.tmu_state2, mu_state_filt)
        mat_vec_mult(self.tmu_state2, var_trans, tmu_alpha2,
                     tmu_beta2, self.tvar_state2, self.tmu_state)
        mat_copy(self.tvar_state3, var_state_filt)
        mat_mult(self.tvar_state3, var_trans, var_trans2, tvar_alpha2, tvar_beta2, self.tvar_state2,
                 self.tvar_state)
        state_sim(x_state_smooth, self.llt_state, self.tmu_state2,
                  self.tvar_state3, z_state)
        return

    cpdef void smooth(self,
                      double[::1] x_state_smooth,
                      double[::1] mu_state_smooth,
                      double[::1, :] var_state_smooth,
                      const double[::1] x_state_next,
                      const double[::1] mu_state_next,
                      const double[::1, :] var_state_next,
                      const double[::1] mu_state_filt,
                      const double[::1, :] var_state_filt,
                      const double[::1] mu_state_pred,
                      const double[::1, :] var_state_pred,
                      const double[::1, :] wgt_state,
                      const double[::1] z_state):
        r"""
        Perform one step of both Kalman mean/variance and sampling smoothers.
        Combines :func:`KalmanTV.smooth_mv` and :func:`KalmanTV.smooth_sim` steps to get :math:`x_{n|N}` and 
        :math:`\theta_{n|N}` from :math:`\theta_{n+1|N}`, :math:`\theta_{n|n}`, and :math:`\theta_{n+1|n}`.
        """
        self.smooth_mv(mu_state_smooth, var_state_smooth,
                       mu_state_next, var_state_next,
                       mu_state_filt, var_state_filt,
                       mu_state_pred, var_state_pred, wgt_state)
        self.smooth_sim(x_state_smooth, x_state_next,
                        mu_state_filt, var_state_filt,
                        mu_state_pred, var_state_pred,
                        wgt_state, z_state)
        return
    
    cpdef void forecast(self,
                        double[::1] mu_fore,
                        double[::1, :] var_fore,
                        const double[::1] mu_state_pred,
                        const double[::1, :] var_state_pred,
                        const double[::1] mu_meas,
                        const double[::1, :] wgt_meas,
                        const double[::1, :] var_meas):
        r"""
        Forecasts the mean and variance of the measurement at time step n given observations from times [0...n-1].
        """
        cdef char * wgt_trans = 'N'
        cdef char * wgt_trans2 = 'T'
        cdef char * var_trans = 'N'
        cdef int mu_alpha = 1, mu_beta = 1, var_alpha = 1, var_beta = 1
        vec_copy(mu_fore, mu_meas)
        mat_vec_mult(mu_fore, wgt_trans, mu_alpha,
                     mu_beta, wgt_meas, mu_state_pred)
        mat_copy(var_fore, var_meas)
        mat_triple_mult(var_fore, self.twgt_meas, wgt_trans, var_trans, wgt_trans2,
                        var_alpha, var_beta, wgt_meas, var_state_pred, wgt_meas)
        
        return



import numpy as np
import kalmantv.numba.scipy_linalg as scipy_linalg
import scipy as sp
import scipy.linalg
from numba import intc, float64
# from numba import jit
from numba.extending import register_jitable
from numba.experimental import jitclass

# --- helper functions ---------------------------------------------------------


@register_jitable
def _solveV(X, V, B, V_chol):
    r"""
    Solves X in :math:`VX = B`, where V is a variance matrix.

    Args:
        X (ndarray(N, M)): Returned matrix.
        V (ndarray(N, N)): Variance matrix.
        B (ndarray(N, M)): Second matrix.
        V_chol (ndarray(N, M)): Temporary matrix in which to store the cholesky factorization of V.

    Returns:
        (tuple):
        - **U** (ndarray(N, M)): Temp matrix.
        - **X** (ndarray(N, M)): X in :math:`VX = B`.

    """
    # no malloc in any of this
    V_chol[:] = V
    X[:] = B
    (V_chol, lower) = scipy.linalg.cho_factor(
        V_chol, overwrite_a=True, check_finite=False)
    X = scipy.linalg.cho_solve(
        (V_chol, lower), b=X, overwrite_b=True, check_finite=False)
    return


@register_jitable
def _quad_form(D, A, B, AB):
    r"""
    Calculates the quadratic form :math:`D += A B A'`.

    Args:
        D (ndarray(N, N)): Returned matrix.
        A (ndarray(N, M)): First matrix.
        B (ndarray(M, M)): Second matrix.
        AB (ndarray(N, M)): Temporary matrix in which to store `AB`.

    Returns:
        (tuple):
        - **D** (ndarray(N, N)): :math:`D = A B A' + D`.
        - **AB** (ndarray(N, M)): :math:`AB`.

    """
    AB[:] = np.dot(A, B)
    D += np.dot(AB, A.T)
    return


@register_jitable
def _mvn_sim(x, mu, V, z, V_chol):
    r"""
    Creates a draw from :math:`x \sim N(\mu, V)` from iid normals :math:`z \sim \N(0, I)`.

    Args:
        x (ndarray(n)): Generated draw.
        mu (ndarray(n)): Mean vector.
        V (ndarray(n, n)): Variance matrix.
        z (ndarray(n)): Random vector simulated from :math:`N(0, 1)`.
        V_chol (ndarray(n, n)): Temporary matrix to store lower Cholesky factor of `V`.

    Returns:
        (tuple):
        - **x** (ndarray(n)): Generated state.
        - **V_chol** (ndarray(n, n)): Lower Cholesky factor of `V`.
    """
    V_chol, lower = scipy.linalg.cho_factor(V, lower=True, check_finite=False)
    x[:] = z
    x = scipy_linalg.tri_mult((V_chol, lower), x,
                              overwrite_x=True, check_finite=False)
    x += mu
    return


_KalmanTV_spec = [
    ('n_meas', intc),
    ('n_state', intc),
    ('tmu_state', float64[::1]),
    ('tmu_state2', float64[::1]),
    ('tvar_state', float64[::1, :]),
    ('tvar_state2', float64[::1, :]),
    ('tvar_state3', float64[::1, :]),
    ('tmu_meas', float64[::1]),
    ('tvar_meas', float64[::1, :]),
    ('twgt_meas', float64[::1, :]),
    ('twgt_meas2', float64[::1, :]),
    ('llt_meas', float64[::1, :]),
    ('llt_state', float64[::1, :])
]


@ jitclass(_KalmanTV_spec)
class KalmanTV(object):
    r"""
    Create a Kalman Time-Varying object. The methods of the object can predict, update, sample and 
    smooth the mean and variance of the Kalman Filter. This method is useful if one wants to track 
    an object with streaming observations.

    The specific model we are using to approximate the solution :math:`x_n` is

    .. math::

        x_n = Q(x_{n-1} -\lambda) + \lambda + R_n^{1/2} \epsilon_n

        y_n = d + W x_n + \Sigma_n^{1/2} \eta_n

    where :math:`\epsilon_n` and :math:`\eta_n` are independent :math:`N(0,1)` distributions and
    :math:`y_n` denotes the model interrogation (observation) at time n.

    The variables of the model are defined below in the argument section. The methods of this class
    calculates :math:`\theta = (\mu, \Sigma)` for :math:`x_n` and the notation for
    the state at time n given observations from k is given by :math:`\theta_{n|K}`.

    Notes:
      - For best performance, all input arrrays should have contiguous memory in fortran-order.
      - Avoids memory allocation whenever possible.  One place this does not happen is in calculations of `A += B C`.  This is done with `A += np.dot(B, C)`, which involves malloc before the addition.

    Args:
        n_state (int): Number of state variables.
        n_meas (int): Number of measurement variables.
        mu_state_past (ndarray(n_state)): Mean estimate for state at time n-1 given observations from 
            times [0...n-1]; :math:`\mu_{n-1|n-1}`. 
        var_state_past (ndarray(n_state, n_state)): Covariance of estimate for state at time n-1 given 
            observations from times [0...n-1]; :math:`\Sigma_{n-1|n-1}`.
        mu_state_pred (ndarray(n_state)): Mean estimate for state at time n given observations from 
            times [0...n-1]; denoted by :math:`\mu_{n|n-1}`. 
        var_state_pred (ndarray(n_state, n_state)): Covariance of estimate for state at time n given 
            observations from times [0...n-1]; denoted by :math:`\Sigma_{n|n-1}`.
        mu_state_filt (ndarray(n_state)): Mean estimate for state at time n given observations from 
            times [0...n]; denoted by :math:`\mu_{n|n}`. 
        var_state_filt (ndarray(n_state, n_state)): Covariance of estimate for state at time n given 
            observations from times [0...n]; denoted by :math:`\Sigma_{n|n}`.
        mu_state_next (ndarray(n_state)): Mean estimate for state at time n+1 given observations from 
            times [0...N]; denoted by :math:`\mu_{n+1|N}`. 
        var_state_next (ndarray(n_state, n_state)): Covariance of estimate for state at time n+1 given 
            observations from times [0...N]; denoted by :math:`\Sigma_{n+1|N}`.
        x_state_smooths (ndarray(n_state)): Sample solution at time n given observations from times [0...N];
            denoted by :math:`X_{n|N}`
        mu_state_smooth (ndarray(n_state)): Mean estimate for state at time n given observations from 
            times [0...N]; denoted by :math:`\mu_{n|N}`.
        var_state_smooth (ndarray(n_state, n_state)): Covariance of estimate for state at time n given 
            observations from times [0...N]; denoted by :math:`\Sigma_{n|N}`.
        x_state (ndarray(n_state)): Simulated state vector; :math:`x_n`.
        mu_state (ndarray(n_state)): Transition offsets defining the solution prior; denoted by :math:`\lambda`.
        wgt_state (ndarray(n_state, n_state)): Transition matrix defining the solution prior; denoted by :math:`Q`.
        var_state (ndarray(n_state, n_state)): Variance matrix defining the solution prior; denoted by :math:`R`.
        x_meas (ndarray(n_meas)): Interrogated measure vector from `x_state`; :math:`y_n`.
        mu_meas (ndarray(n_meas)): Transition offsets defining the measure prior; denoted by :math:`d`.
        wgt_meas (ndarray(n_meas, n_state)): Transition matrix defining the measure prior; denoted by :math:`W`.
        var_meas (ndarray(n_meas, n_meas)): Variance matrix defining the measure prior; denoted by :math:`\Sigma_n`.
        z_state (ndarray(n_state)): Random vector simulated from :math:`N(0, 1)`.
        mu_fore (ndarray(n_meas)): Mean estimate for measurement at n given observations from [0...n-1]
        var_fore (ndarray(n_meas, n_meas)): Covariance of estimate for state at time n given 
            observations from times [0...n-1]

    """

    def __init__(self, n_meas: int, n_state: int) -> None:
        self.n_meas = n_meas
        self.n_state = n_state
        self.tmu_state = np.empty(n_state)
        self.tmu_state2 = np.empty(n_state)
        # note: Numba doesn't support `order='F'`, so emulate
        # via array transpose as documented here: https://numba.pydata.org/numba-doc/dev/user/faq.html#how-can-i-create-a-fortran-ordered-array
        self.tvar_state = np.empty((n_state, n_state)).T
        self.tvar_state2 = np.empty((n_state, n_state)).T
        self.tvar_state3 = np.empty((n_state, n_state)).T
        self.tmu_meas = np.empty(n_meas)
        self.tvar_meas = np.empty((n_meas, n_meas)).T
        self.twgt_meas = np.empty((n_state, n_meas)).T
        self.twgt_meas2 = np.empty((n_state, n_meas)).T
        self.llt_meas = np.empty((n_meas, n_meas)).T
        self.llt_state = np.empty((n_state, n_state)).T

    def predict(self,
                mu_state_pred,
                var_state_pred,
                mu_state_past,
                var_state_past,
                mu_state,
                wgt_state,
                var_state):
        r"""
        Perform one prediction step of the Kalman filter.
        Calculates :math:`\theta_{n|n-1}` from :math:`\theta_{n-1|n-1}`.

        Note: `mu_state_pred` and `mu_state_past` cannot refer to the same location in memory.  Same for `var_state_pred` and `var_state_past`.
        """
        mu_state_pred[:] = mu_state
        mu_state_pred += np.dot(wgt_state, mu_state_past)
        var_state_pred[:] = var_state
        _quad_form(var_state_pred, wgt_state, var_state_past, self.tvar_state)

        return

    def update(self,
               mu_state_filt,
               var_state_filt,
               mu_state_pred,
               var_state_pred,
               x_meas,
               mu_meas,
               wgt_meas,
               var_meas):
        r"""
        Perform one update step of the Kalman filter.
        Calculates :math:`\theta_{n|n}` from :math:`\theta_{n|n-1}`.

        Note: `mu_state_filt` and `mu_state_pred` can refer to the same location in memory.  Same for `var_state_filt` and `var_state_pred`.
        """
        self.tmu_meas[:] = -mu_meas
        self.tmu_meas -= np.dot(wgt_meas, mu_state_pred)
        self.tvar_meas[:] = var_meas
        _quad_form(self.tvar_meas, wgt_meas, var_state_pred, self.twgt_meas)
        _solveV(self.twgt_meas2, self.tvar_meas, self.twgt_meas, self.llt_meas)
        self.tmu_meas += x_meas
        mu_state_filt[:] = mu_state_pred
        mu_state_filt += np.dot(self.twgt_meas2.T, self.tmu_meas)
        var_state_filt[:] = var_state_pred
        var_state_filt -= np.dot(self.twgt_meas2.T, self.twgt_meas)
        return

    def filter(self,
               mu_state_pred,
               var_state_pred,
               mu_state_filt,
               var_state_filt,
               mu_state_past,
               var_state_past,
               mu_state,
               wgt_state,
               var_state,
               x_meas,
               mu_meas,
               wgt_meas,
               var_meas):
        r"""
        Perform one step of the Kalman filter.
        Combines :func:`KalmanTV.predict` and :func:`KalmanTV.update` steps to get :math:`\theta_{n|n}` from :math:`\theta_{n-1|n-1}`.
        """
        self.predict(mu_state_pred, var_state_pred,
                     mu_state_past, var_state_past,
                     mu_state, wgt_state, var_state)
        self.update(mu_state_filt, var_state_filt,
                    mu_state_pred, var_state_pred,
                    x_meas, mu_meas, wgt_meas, var_meas)
        return

    def smooth_mv(self,
                  mu_state_smooth,
                  var_state_smooth,
                  mu_state_next,
                  var_state_next,
                  mu_state_filt,
                  var_state_filt,
                  mu_state_pred,
                  var_state_pred,
                  wgt_state):
        r"""
        Perform one step of the Kalman mean/variance smoother.
        Calculates :math:`\theta_{n|N}` from :math:`\theta_{n+1|N}`, :math:`\theta_{n|n}`, and :math:`\theta_{n+1|n}`.

        Note: `mu_state_smooth` and `mu_state_next` can refer to the same location in memory.  Same for `var_state_smooth` and `var_state_next`.
        """
        self.tvar_state[:] = np.dot(wgt_state, var_state_filt.T)
        _solveV(self.tvar_state, var_state_pred,
                self.tvar_state, self.llt_state)
        self.tmu_state[:] = mu_state_next - mu_state_pred
        self.tvar_state2[:] = var_state_next - var_state_pred
        self.tvar_state3[:] = np.dot(self.tvar_state.T, self.tvar_state2)
        mu_state_smooth[:] = mu_state_filt
        mu_state_smooth += np.dot(self.tvar_state.T, self.tmu_state)
        var_state_smooth[:] = var_state_filt
        var_state_smooth += np.dot(self.tvar_state3, self.tvar_state)
        return

    def smooth_sim(self,
                   x_state_smooth,
                   x_state_next,
                   mu_state_filt,
                   var_state_filt,
                   mu_state_pred,
                   var_state_pred,
                   wgt_state,
                   z_state):
        r"""
        Perform one step of the Kalman sampling smoother.
        Calculates a draw :math:`x_{n|N}` from :math:`x_{n+1|N}`, :math:`\theta_{n|n}`, and :math:`\theta_{n+1|n}`.

        Note: `x_state_smooth` and `x_state_next` can refer to the same location in memory.
        """
        self.tvar_state[:] = np.dot(wgt_state, var_state_filt.T)
        _solveV(self.tvar_state2, var_state_pred,
                self.tvar_state, self.llt_state)
        self.tmu_state[:] = x_state_next - mu_state_pred
        self.tmu_state2[:] = mu_state_filt
        self.tmu_state2 += np.dot(self.tvar_state2.T, self.tmu_state)
        self.tvar_state3[:] = var_state_filt
        self.tvar_state3 -= np.dot(self.tvar_state2.T, self.tvar_state)
        _mvn_sim(x_state_smooth, self.tmu_state2, self.tvar_state3,
                 z_state, self.llt_state)
        return

    def smooth(self,
               x_state_smooth,
               mu_state_smooth,
               var_state_smooth,
               x_state_next,
               mu_state_next,
               var_state_next,
               mu_state_filt,
               var_state_filt,
               mu_state_pred,
               var_state_pred,
               wgt_state,
               z_state):
        r"""
        Perform one step of both Kalman mean/variance and sampling smoothers.
        Combines :func:`KalmanTV.smooth_mv` and :func:`KalmanTV.smooth_sim` steps to get :math:`x_{n|N}` and
        :math:`\theta_{n|N}` from :math:`\theta_{n+1|N}`, :math:`\theta_{n|n}`, and :math:`\theta_{n+1|n}`.
        """
        self.smooth_mv(mu_state_smooth, var_state_smooth,
                       mu_state_next, var_state_next,
                       mu_state_filt, var_state_filt,
                       mu_state_pred, var_state_pred, wgt_state)
        self.smooth_sim(x_state_smooth, x_state_next,
                        mu_state_filt, var_state_filt,
                        mu_state_pred, var_state_pred,
                        wgt_state, z_state)
        return

    def forecast(self,
                 mu_fore,
                 var_fore,
                 mu_state_pred,
                 var_state_pred,
                 mu_meas,
                 wgt_meas,
                 var_meas):
        r"""
        Forecasts the mean and variance of the measurement at time step n given observations from times [0...n-1].
        """
        mu_fore[:] = mu_meas
        mu_fore += np.dot(wgt_meas, mu_state_pred)
        var_fore[:] = var_meas
        _quad_form(var_fore, wgt_meas, var_state_pred, self.twgt_meas)
        return



from KalmanTV cimport KalmanTV_raw as CKalmanTV

cdef class _KalmanTV:
    r"""
    Create a Kalman Time-Varying object. The methods of the object can predict, update, sample and 
    smooth the mean and variance of the Kalman Filter. This method is useful if one wants to track 
    an object with streaming observations.

    The specific model we are using to approximate the solution :math:`x_n` is

    .. math::

        x_n = Q(x_{n-1} -\lambda) + \lambda + R_n^{1/2} \epsilon_n

        y_n = d + W x_n + \Sigma_n^{1/2} \eta_n

    where :math:`\epsilon_n` and :math:`\eta_n` are independent :math:`N(0,1)` distributions and
    :math:`y_n` denotes the model interrogation (observation) at time n.

    The variables of the model are defined below in the argument section. The methods of this class
    calculates :math:`\theta = (\mu, \Sigma)` for :math:`x_n` and the notation for
    the state at time n given observations from k is given by :math:`\theta_{n|K}`.

    Args:
        n_state (int): Number of state variables.
        n_meas (int): Number of measurement variables.
        mu_state_past (ndarray(n_state)): Mean estimate for state at time n-1 given observations from 
            times [0...n-1]; :math:`\mu_{n-1|n-1}`. 
        var_state_past (ndarray(n_state, n_state)): Covariance of estimate for state at time n-1 given 
            observations from times [0...n-1]; :math:`\Sigma_{n-1|n-1}`.
        mu_state_pred (ndarray(n_state)): Mean estimate for state at time n given observations from 
            times [0...n-1]; denoted by :math:`\mu_{n|n-1}`. 
        var_state_pred (ndarray(n_state, n_state)): Covariance of estimate for state at time n given 
            observations from times [0...n-1]; denoted by :math:`\Sigma_{n|n-1}`.
        mu_state_filt (ndarray(n_state)): Mean estimate for state at time n given observations from 
            times [0...n]; denoted by :math:`\mu_{n|n}`. 
        var_state_filt (ndarray(n_state, n_state)): Covariance of estimate for state at time n given 
            observations from times [0...n]; denoted by :math:`\Sigma_{n|n}`.
        mu_state_next (ndarray(n_state)): Mean estimate for state at time n+1 given observations from 
            times [0...N]; denoted by :math:`\mu_{n+1|N}`. 
        var_state_next (ndarray(n_state, n_state)): Covariance of estimate for state at time n+1 given 
            observations from times [0...N]; denoted by :math:`\Sigma_{n+1|N}`.
        x_state_smooths (ndarray(n_state)): Sample solution at time n given observations from times [0...N];
            denoted by :math:`X_{n|N}`
        mu_state_smooth (ndarray(n_state)): Mean estimate for state at time n given observations from 
            times [0...N]; denoted by :math:`\mu_{n|N}`.
        var_state_smooth (ndarray(n_state, n_state)): Covariance of estimate for state at time n given 
            observations from times [0...N]; denoted by :math:`\Sigma_{n|N}`.
        x_state (ndarray(n_state)): Simulated state vector; :math:`x_n`.
        mu_state (ndarray(n_state)): Transition offsets defining the solution prior; denoted by :math:`\lambda`.
        wgt_state (ndarray(n_state, n_state)): Transition matrix defining the solution prior; denoted by :math:`Q`.
        var_state (ndarray(n_state, n_state)): Variance matrix defining the solution prior; denoted by :math:`R`.
        x_meas (ndarray(n_meas)): Interrogated measure vector from `x_state`; :math:`y_n`.
        mu_meas (ndarray(n_meas)): Transition offsets defining the measure prior; denoted by :math:`d`.
        wgt_meas (ndarray(n_meas, n_state)): Transition matrix defining the measure prior; denoted by :math:`W`.
        var_meas (ndarray(n_meas, n_meas)): Variance matrix defining the measure prior; denoted by :math:`\Sigma_n`.
        z_state (ndarray(n_state)): Random vector simulated from :math:`N(0, 1)`.
        mu_fore (ndarray(n_meas)): Mean estimate for measurement at n given observations from [0...n-1]
        var_fore (ndarray(n_meas, n_meas)): Covariance of estimate for state at time n given 
            observations from times [0...n-1]

    """
    cdef CKalmanTV * ktv

    def __cinit__(self, int n_meas, int n_state):
        self.ktv = new CKalmanTV(n_meas, n_state)

    def __dealloc__(self):
        del self.ktv

    def predict(self,
                double[::1] mu_state_pred,
                double[::1, :] var_state_pred,
                const double[::1] mu_state_past,
                const double[::1, :] var_state_past,
                const double[::1] mu_state,
                const double[::1, :] wgt_state,
                const double[::1, :] var_state):
        """
        Perform one prediction step of the Kalman filter.
        Calculates :math:`\\theta_{n|n-1}` from :math:`\\theta_{n-1|n-1}`.
        """
        self.ktv.predict(& mu_state_pred[0], & var_state_pred[0, 0],
                         & mu_state_past[0], & var_state_past[0, 0],
                         & mu_state[0], & wgt_state[0, 0], & var_state[0, 0])
        return

    def update(self,
               double[::1] mu_state_filt,
               double[::1, :] var_state_filt,
               const double[::1] mu_state_pred,
               const double[::1, :] var_state_pred,
               const double[::1] x_meas,
               const double[::1] mu_meas,
               const double[::1, :] wgt_meas,
               const double[::1, :] var_meas):
        """
        Perform one update step of the Kalman filter.
        Calculates :math:`\\theta_{n|n}` from :math:`\\theta_{n|n-1}`.
        """
        self.ktv.update(& mu_state_filt[0], & var_state_filt[0, 0],
                        & mu_state_pred[0], & var_state_pred[0, 0],
                        & x_meas[0], & mu_meas[0],
                        & wgt_meas[0, 0], & var_meas[0, 0])
        return

    def filter(self,
               double[::1] mu_state_pred,
               double[::1, :] var_state_pred,
               double[::1] mu_state_filt,
               double[::1, :] var_state_filt,
               const double[::1] mu_state_past,
               const double[::1, :] var_state_past,
               const double[::1] mu_state,
               const double[::1, :] wgt_state,
               const double[::1, :] var_state,
               const double[::1] x_meas,
               const double[::1] mu_meas,
               const double[::1, :] wgt_meas,
               const double[::1, :] var_meas):
        """
        Perform one step of the Kalman filter.
        Combines :func:`KalmanTV.predict` and :func:`KalmanTV.update` steps to get :math:`\\theta_{n|n}` from :math:`\\theta_{n-1|n-1}`.
        """
        self.ktv.filter(& mu_state_pred[0], & var_state_pred[0, 0],
                        & mu_state_filt[0], & var_state_filt[0, 0],
                        & mu_state_past[0], & var_state_past[0, 0],
                        & mu_state[0], & wgt_state[0, 0], & var_state[0, 0],
                        & x_meas[0], & mu_meas[0],
                        & wgt_meas[0, 0], & var_meas[0, 0])
        return

    def smooth_mv(self,
                  double[::1] mu_state_smooth,
                  double[::1, :] var_state_smooth,
                  const double[::1] mu_state_next,
                  const double[::1, :] var_state_next,
                  const double[::1] mu_state_filt,
                  const double[::1, :] var_state_filt,
                  const double[::1] mu_state_pred,
                  const double[::1, :] var_state_pred,
                  const double[::1, :] wgt_state):
        """
        Perform one step of the Kalman mean/variance smoother.
        Calculates :math:`\\theta_{n|N}` from :math:`\\theta_{n+1|N}`, :math:`\\theta_{n|n}`, and :math:`\\theta_{n+1|n}`.
        """
        self.ktv.smooth_mv(& mu_state_smooth[0], & var_state_smooth[0, 0],
                           & mu_state_next[0], & var_state_next[0, 0],
                           & mu_state_filt[0], & var_state_filt[0, 0],
                           & mu_state_pred[0], & var_state_pred[0, 0],
                           & wgt_state[0, 0])
        return

    def smooth_sim(self,
                   double[::1] x_state_smooth,
                   const double[::1] x_state_next,
                   const double[::1] mu_state_filt,
                   const double[::1, :] var_state_filt,
                   const double[::1] mu_state_pred,
                   const double[::1, :] var_state_pred,
                   const double[::1, :] wgt_state,
                   const double[::1] z_state):
        """
        Perform one step of the Kalman sampling smoother.
        Calculates a draw :math:`x_{n|N}` from :math:`x_{n+1|N}`, :math:`\\theta_{n|n}`, and :math:`\\theta_{n+1|n}`.
        """
        self.ktv.smooth_sim(& x_state_smooth[0], & x_state_next[0],
                            & mu_state_filt[0], & var_state_filt[0, 0],
                            & mu_state_pred[0], & var_state_pred[0, 0],
                            & wgt_state[0, 0], & z_state[0])
        return

    def smooth(self,
               double[::1] x_state_smooth,
               double[::1] mu_state_smooth,
               double[::1, :] var_state_smooth,
               const double[::1] x_state_next,
               const double[::1] mu_state_next,
               const double[::1, :] var_state_next,
               const double[::1] mu_state_filt,
               const double[::1, :] var_state_filt,
               const double[::1] mu_state_pred,
               const double[::1, :] var_state_pred,
               const double[::1, :] wgt_state,
               const double[::1] z_state):
        """
        Perform one step of both Kalman mean/variance and sampling smoothers.
        Combines :func:`KalmanTV.smooth_mv` and :func:`KalmanTV.smooth_sim` steps to get :math:`x_{n|N}` and 
        :math:`\\theta_{n|N}` from :math:`\\theta_{n+1|N}`, :math:`\\theta_{n|n}`, and :math:`\\theta_{n+1|n}`.
        """
        self.ktv.smooth(& x_state_smooth[0], & mu_state_smooth[0],
                        & var_state_smooth[0, 0], & x_state_next[0],
                        & mu_state_next[0], & var_state_next[0, 0],
                        & mu_state_filt[0], & var_state_filt[0, 0],
                        & mu_state_pred[0], & var_state_pred[0, 0],
                        & wgt_state[0, 0], & z_state[0])
        return

    def state_sim(self,
                  double[::1] x_state,
                  const double[::1] mu_state,
                  const double[::1, :] var_state,
                  const double[::1] z_state):
        """
        Simulates from a normal distribution with mean `mu_state`, variance `var_state`,
        and randomness `z_state` drawn from :math:`N(0, 1)`.
        """
        self.ktv.state_sim(& x_state[0], & mu_state[0],
                            & var_state[0, 0], & z_state[0])
        return
        
    def forecast(self,
                 double[::1] mu_fore,
                 double[::1, :] var_fore,
                 const double[::1] mu_state_pred,
                 const double[::1, :] var_state_pred,
                 const double[::1] mu_meas,
                 const double[::1, :] wgt_meas,
                 const double[::1, :] var_meas):
        r"""
        Forecasts the mean and variance of the measurement at time step n given observations from times [0...n-1].
        """
        self.ktv.forecast(& mu_fore[0], & var_fore[0, 0],
                          & mu_state_pred[0], & var_state_pred[0, 0],
                          & mu_meas[0], & wgt_meas[0, 0], & var_meas[0, 0])
        return



import numpy as np

def ss2genss(wgt_state, mu_state, var_state, wgt_meas, mu_meas, var_meas):
    r"""
    Converts the state-space parameters:

    .. math::

        X_n = c_n + T_n X_{n-1} + R_n^{1/2} \epsilon_n

        y_n = d_n + W_n x_n + H_n^{1/2} \eta_n

    to the Gaussian process parameters with the model:

    .. math::

        Y_n = b_n + A_n Y_{n-1} + C_n \epsilon_n

    Args:
        wgt_state (ndarray(n_states, n_states, n_steps)): Transition matrices defining
            the solution prior; denoted by :math:`T_n`.
        mu_state (ndarray(n_state, n_steps)): Transition offsets defining the solution 
            prior; denoted by :math:`c_n`.
        var_state (ndarray(n_state, n_state, n_steps)): Variance matrices
            defining the solution prior; denoted by :math:`R_n`.
        wgt_meas (ndarray(n_meas, n_meas, n_steps)): Transition matrices defining the 
            measure prior; denoted by :math:`W_n`.
        mu_meas (ndarray(n_meas, n_steps)): Transition offsets defining the measure prior; 
            denoted by :math:`d_n`.
        var_meas (ndarray(n_meas, n_meas, n_steps)): Variance matrice defining 
            the measure prior; denoted by :math:`H_n`.

    Returns:
        (tuple):
        - **wgt_genss** (ndarray(n_genss, n_genss, steps)): Transition matrices in the Gaussian process; denoted by :math:`A_n`.
        - **mu_genss** (ndarray(n_genss, steps)): Transition offsets in the Gaussian process; denoted by :math:`b_n`.
        - **chol_genss** (ndarray(n_genss, n_genss, steps)): Cholesky of the variance matrices in the Gaussian process; denoted by :math:`C_n`.

    """
    # dimensions
    n_state, n_steps = mu_state.shape
    n_meas = mu_meas.shape[0]
    n_genss = n_state + n_meas

    # initialize genss variables
    wgt_genss = np.zeros((n_genss, n_genss, n_steps), order='F')
    mu_genss = np.zeros((n_genss, n_steps), order='F')
    chol_genss = np.zeros((n_genss, n_genss, n_steps), order='F')

    for i in range(n_steps):
        # convert var to chol
        chol_state = np.linalg.cholesky(var_state[:, :, i])
        chol_meas = np.linalg.cholesky(var_meas[:, :, i])

        # wgt_genss
        wgt = np.zeros((n_genss, n_genss), order='F')
        wgt[0:n_state, 0:n_state] = wgt_state[:, :, i]
        wgt[n_state:, 0:n_state] = wgt_meas[:, :, i].dot(wgt_state[:, :, i])
        wgt_genss[:, :, i] = wgt

        # mu_genss
        mu = np.zeros(n_genss)
        mu[0:n_state] = mu_state[:, i]
        mu[n_state:] = wgt_meas[:, :, i].dot(mu_state[:, i]) + mu_meas[:, i]
        mu_genss[:, i] = mu

        # chol_genss
        chol = np.zeros((n_genss, n_genss), order='F')
        chol[0:n_state, 0:n_state] = chol_state
        chol[n_state:, 0:n_state] = wgt_meas[:, :, i].dot(chol_state)
        chol[n_state:, n_state:] = chol_meas
        chol_genss[:, :, i] = chol

    wgt_genss[:, :, 0] = np.nan
    return wgt_genss, mu_genss, chol_genss


def mv_gaussian(wgt_genss, mu_genss, chol_genss):
    r"""
    Calculate the mean and variance of the joint density :math:`Y=(Y_0, Y_1, \ldots, Y_n)` where 
    :math:`Y_n` is the Gaussian process.

    Args:
        wgt_genss (ndarray(n_genss, n_genss, steps)): Transition matrices in the Gaussian process;
            denoted by :math:`A_n`.
        mu_genss (ndarray(n_genss, steps)): Transition offsets in the Gaussian process;
            denoted by :math:`b_n`.
        chol_genss (ndarray(n_genss, n_genss, steps)): Cholesky of the variance matrices in the 
            Gaussian process; denoted by :math:`C_n`.

    Returns:
        (tuple):
        - **gaussian_mu** (ndarray(n_genss, n_genss, steps)): Mean of the joint density.
        - **gaussian_var** (ndarray(n_genss, steps)): Variance of the joint density.

    """
    n_genss, n_steps = mu_genss.shape
    D = n_genss*n_steps
    An_m = np.zeros((n_genss, n_genss, n_steps, n_steps+1))
    for n in range(n_steps):
        for m in range(n_steps+1):
            if m > n:
                An_m[:, :, n, m] = np.eye(n_genss)
            elif n == m:
                An_m[:, :, n, m] = wgt_genss[:, :, n]
            else:
                diff = n-m
                wgt_diff = wgt_genss[:, :, m]
                for i in range(diff):
                    wgt_diff = np.matmul(wgt_genss[:, :, m+i+1], wgt_diff)
                An_m[:, :, n, m] = wgt_diff

    L = np.zeros((D, D))
    gaussian_mu = np.zeros(D)
    for n in range(n_steps):
        for m in range(n, n_steps):
            L[m*n_genss:(m+1)*n_genss, n*n_genss:(n+1) *
              n_genss] = np.matmul(An_m[:, :, m, n+1], chol_genss[:, :, n])
        for m in range(n+1):
            gaussian_mu[n*n_genss:(n+1)*n_genss] = gaussian_mu[n*n_genss:(n+1)
                                                           * n_genss] + An_m[:, :, n, m+1].dot(mu_genss[:, m])
    gaussian_var = L.dot(L.T)
    return gaussian_mu, gaussian_var



# cython: boundscheck=False, wraparound=False, nonecheck=False, initializedcheck=False
import numpy as np
cimport numpy as np
cimport cython
cimport scipy.linalg.cython_blas as blas
from scipy.linalg.cython_lapack cimport dpotrf, dpotrs, dlacpy

cpdef void vec_copy(double[::1] y,
                    const double[::1] x):
    r"""
    Copies vector x to y.

    Args:
        y (ndarray(N)): Returned vector.
        x (ndarray(N)): Vector x.

    Returns:
        (ndarray(M)): Copied vector y.

    """
    cdef int N = len(x), incx = 1, incy = 1
    blas.dcopy( & N, & x[0], & incx, & y[0], & incy)
    return

cpdef void mat_copy(double[::1, :] B,
                    const double[::1, :] A):
    r"""
    Copies Matrix A to B.

    Args:
        B (ndarray(N)): Returned matrix.
        A (ndarray(N)): Matrix A.

    Returns:
        (ndarray(M)): Copied matrix B.

    """
    cdef char * uplo = 'A'
    cdef int M = A.shape[0], N = A.shape[1], lda = M, ldb = M
    dlacpy( & uplo[0], & M, & N, & A[0, 0], & lda, & B[0, 0], & ldb)
    return

cpdef void vec_add(double[::1] y,
                   const double alpha,
                   const double[::1] x):
    r"""
    Calculates :math:`y = \alpha x + y`.

    Args:
        y (ndarray(N)): Returned vector.
        alpha (double): Scalar :math:`\alpha`.
        x (ndarray(N)): Vector x.

    Returns:
        (ndarray(M)): :math:`y = \alpha x + y`.

    """
    cdef int N = len(x), incx = 1, incy = 1
    blas.daxpy(& N, & alpha, & x[0], & incx, & y[0], & incy)
    return

cpdef void mat_add(double[::1, :] B,
                   const double alpha,
                   const double beta,
                   const double[::1, :] A):
    r"""
    Calculates :math:`B = \alpha A + \beta B`.

    Args:
        B (ndarray(N)): Returned matrix.
        alpha (double): Scalar :math:`\alpha`.
        beta (double): Scalar :math:`\beta`.
        A (ndarray(N)): Matrix A.
        
    Returns:
        (ndarray(M)): :math:`B = \alpha A + \beta B`.

    """
    cdef int M = A.shape[0], N = A.shape[1]
    for i in range(M):
        for j in range(N):
            B[i, j] = alpha*A[i, j] + beta*B[i, j]
    return

cpdef void mat_vec_mult(double[::1] y,
                        char * trans,
                        const double alpha,
                        const double beta,
                        const double[::1, :] A,
                        const double[::1] x):
    r"""
    Calculates :math:`y = \alpha A x + \beta y`.

    Args:
        y (ndarray(M)): Returned vector.
        trans (char): Specifies if matrix A should be transposed.
        alpha (double): Scalar :math:`\alpha`.
        beta (double): Scalar :math:`\beta`.
        A (ndarray(M, N)): Matrix A.
        x (ndarray(N)): Vector x.
        
    Returns:
        (ndarray(M)): :math:`y = \alpha A x + \beta y`.

    """
    cdef int M = A.shape[0], N = A.shape[1], lda = M, incx = 1, incy = 1
    blas.dgemv( & trans[0], & M, & N, & alpha, & A[0, 0], & lda, & x[0], & incx, & beta, & y[0], & incy)
    return

cpdef void tri_vec_mult(double[::1] x,
                        char * uplo,
                        char * trans,
                        char * diag,
                        const double[::1, :] A):
    r"""
    Calculates :math:`x = A x` where A is a triangular matrix.

    Args:
        x (ndarray(N)): Vector x.
        uplo (char): Specifies if matrix A is upper or lower triangular.
        trans (char): Specifies if matrix A should be transposed.
        diag (char): Specifies if matrix A is unit triangular.
        A (ndarray(M, N)): Matrix A.

    Returns:
        (ndarray(M)): :math:`x = A x`.

    """
    cdef int N = A.shape[0], lda = N, incx = 1
    blas.dtrmv(& uplo[0], & trans[0], & diag[0], & N, & A[0, 0], & lda, & x[0], & incx)
    return

cpdef void mat_mult(double[::1, :] C,
                    char * transa,
                    char * transb,
                    const double alpha,
                    const double beta,
                    const double[::1, :] A,
                    const double[::1, :] B):
    r"""
    Calculates :math:`C = \alpha A B + \beta C`.

    Args:
        C (ndarray(M, N)): Returned matrix.
        transa (char): Specifies if matrix A should be transposed.
        transb (char): Specifies if matrix B should be transposed.
        alpha (double): Scalar :math:`\alpha`.
        beta (double): Scalar :math:`\beta`.
        A (ndarray(M, K)): First matrix.
        B (ndarray(K, N)): Second matrix.

    Returns:
        (ndarray(M, N)): :math:`C = \alpha A B + \beta C`.

    """
    # get dimensions
    cdef int M, N, K, lda, ldb, ldc
    if transa == b'N':
        M = A.shape[0]
        K = A.shape[1]
        lda = M
    else:
        M = A.shape[1]
        K = A.shape[0]
        lda = K

    if transb == b'N':
        N = B.shape[1]
        ldb = K
    else:
        N = B.shape[0]
        ldb = N
    ldc = M
    blas.dgemm( & transa[0], & transb[0], & M, & N, & K, & alpha, & A[0, 0], & lda, & B[0, 0], & ldb, & beta, & C[0, 0], & ldc)
    return

cpdef void mat_triple_mult(double[::1, :] D,
                           double[::1, :] temp,
                           char * transa,
                           char * transb,
                           char * transc,
                           const double alpha,
                           const double beta,
                           const double[::1, :] A,
                           const double[::1, :] B,
                           const double[::1, :] C):
    r"""
    Calculates :math:`D = \alpha A B C + \beta D`.

    Args:
        D (ndarray(L, N)): Returned matrix.
        temp (ndarray(M, L)): Temp matrix for intermediate matrix multiplication; :math:`AB`.
        transa (char): Specifies if matrix A should be transposed.
        transb (char): Specifies if matrix B should be transposed.
        transc (char): Specifies if matrix C should be transposed.
        alpha (double): Scalar :math:`\alpha`.
        beta (double): Scalar :math:`\beta`.
        A (ndarray(M, K)): First matrix.
        B (ndarray(K, L)): Second matrix.
        C (ndarray(L, N)): Third matrix.

    Returns:
        (tuple):
        - **D** (ndarray(M, N)): :math:`D = \alpha A B C + \beta D`.
        - **temp** (ndarray(M, L)): :math:`AB`.

    """
    # Temp alpha, beta
    cdef int alpha1 = 1, beta1 = 0
    # Temp trans
    cdef char * trans1 = 'N'
    mat_mult(temp, transa, transb, alpha1, beta1, A, B)
    mat_mult(D, trans1, transc, alpha, beta, temp, C)
    return

cpdef void chol_fact(double[::1, :] L,
                     const double[::1, :] V):
    r"""
    Computes the cholesky factorization of variance matrix V.

    Args:
        L (ndarray(N, M)): Returned matrix.
        V (ndarray(N, N)): Variance matrix.
    
    Returns:
        (ndarray(N, M)): Cholesky factorization of variance matrix V.

    """
    cdef char * uplo = 'L'
    cdef int N = V.shape[0], lda = N, info
    mat_copy(L, V)  # operates in-place, so this prevents V from being overwritten
    dpotrf(& uplo[0], & N, & L[0, 0], & lda, & info)
    return

cpdef void solveV(double[::1, :] U,
                  double[::1, :] X,
                  const double[::1, :] V,
                  const double[::1, :] B):
    r"""
    Solves X in :math:`VX = B`, where V is a variance matrix.

    Args:
        U (ndarray(N, M)): Temp matrix for intermediate storage; cholesky factorization of V.
        X (ndarray(N, M)): Returned matrix.
        V (ndarray(N, N)): Variance matrix.
        B (ndarray(N, M)): Second matrix.
        
    Returns:
        (tuple):
        - **U** (ndarray(N, M)): Temp matrix.
        - **X** (ndarray(N, M)): X in :math:`VX = B`.

    """
    # get dimensions
    cdef int n = V.shape[0], nrhs = B.shape[1], lda = n, ldb = n, info
    cdef char * uplo = 'U'
    mat_copy(U, V)  # operates in-place, so this prevents V from being overwritten
    dpotrf(& uplo[0], & n, & U[0, 0], & lda, & info)
    # solve system with cholesky factor
    mat_copy(X, B)  # again prevents overwriting
    dpotrs(& uplo[0], & n, & nrhs, & U[0, 0], & lda, & X[0, 0], & ldb, & info)
    return



import os
from .omp_init import _omp_init
from .kalmantv import _KalmanTV as KalmanTV


def get_include():
    r"""
    Return the directory that contains the `KalmanTV.h` header file.

    Similar to `numpy.get_include()`, except only works after kalmantv has been installed.
    """
    return os.path.dirname(__file__)
    # import kalmantv
    # if numpy.show_config is None:
    #     # running from numpy source directory
    #     d = os.path.join(os.path.dirname(numpy.__file__), 'core', 'include')
    # else:
    #     # using installed numpy core headers
    #     import numpy.core as core
    #     d = os.path.join(os.path.dirname(core.__file__), 'include')
    # return d



from setuptools import setup, find_packages, Extension
import numpy as np
import platform
import os
import eigenpip as epip
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()
docs_url = "https://kalmantv.readthedocs.io/en/latest/"

# compile with cython if it's installed
try:
    from Cython.Distutils import build_ext
    USE_CYTHON = True
except ImportError:
    USE_CYTHON = False

cmdclass = {}
if USE_CYTHON:
    # extensions = cythonize(extensions)
    cmdclass.update({"build_ext": build_ext})

# readthedocs install 
on_rtd = os.environ.get('READTHEDOCS') == 'True'

# path to eigen library
#EIGEN_PATH = epip.get_include()

#def package_files(directory):
#    paths = []
#    for (path, _, filenames) in os.walk(directory):
#        for filename in filenames:
#            paths.append(os.path.join('..', path, filename))
#    return paths


#extra_files = package_files(EIGEN_PATH)

# compiler options
if platform.system() != "Windows":
    extra_compile_args = ["-O3", "-ffast-math",
                          "-mtune=native", "-march=native"]
    if platform.system() != "Darwin" and not on_rtd:
        # default compiler on macOS doesn't support openmp
        extra_compile_args.append("-fopenmp")
else:
    extra_compile_args = ["-O2", "/openmp"]

# remove numpy depreciation warnings as documented here:
# https://cython.readthedocs.io/en/latest/src/userguide/migrating_to_cy30.html#numpy-c-api
disable_numpy_warnings = [("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]

# cpp modules
ext_c = ".pyx" if USE_CYTHON else ".c"
ext_cpp = ".pyx" if USE_CYTHON else "cpp"
ext_modules = [Extension("kalmantv.cython.blas",
                         ["kalmantv/cython/blas"+ext_c],
                         include_dirs=[
                             np.get_include()],
                         # sp.get_include()],
                         extra_compile_args=extra_compile_args,
                         define_macros=disable_numpy_warnings,
                         language="c"),
               Extension("kalmantv.cython.kalmantv",
                         ["kalmantv/cython/kalmantv"+ext_c],
                         include_dirs=[
                             np.get_include()],
                         extra_compile_args=extra_compile_args,
                         define_macros=disable_numpy_warnings,
                         language="c"),
               Extension("kalmantv.eigen.kalmantv",
                         ["kalmantv/eigen/kalmantv"+ext_cpp],
                         include_dirs=[
                             np.get_include(),
                             epip.get_include()],
                         extra_compile_args=extra_compile_args,
                         define_macros=disable_numpy_warnings,
                         language="c++"),
               Extension("kalmantv.eigen.omp_init",
                         ["kalmantv/eigen/omp_init"+ext_c],
                         include_dirs=[
                             np.get_include()],
                         extra_compile_args=extra_compile_args,
                         define_macros=disable_numpy_warnings,
                         language="c")]

setup(
    name="kalmantv",
    version="0.2.1",
    author="Mohan Wu, Martin Lysy",
    author_email="mlysy@uwaterloo.ca",
    description="High-Performance Kalman Filtering and Smoothing",
    long_description= long_description,
    long_description_content_type='text/markdown',
    keywords="Kalman Cython",
    url="http://github.com/mlysy/kalmantv",
    project_urls = {
        "Documentation": docs_url
    },
    packages=["kalmantv/cython", "kalmantv/numba", "kalmantv/eigen",
              "kalmantv", 
              #"kalmantv/include/eigen"
              ],
    #package_dir={"kalmantv/include/eigen": EIGEN_PATH},
    package_data={
        "kalmantv/cython": ["*.pxd"],
        "kalmantv/eigen": ["*.pxd", "*.h"],
        #"kalmantv/include/eigen": extra_files
    },
    #package_data = packagefiles,
    # cython
    cmdclass=cmdclass,
    ext_modules=ext_modules,

    install_requires=[
        'numpy>=1.16.4', 'scipy>=1.2.1',
        'numba>=0.51.2', 'Cython>=0.29.12',
        'eigenpip'
    ],
    extras_require={
        'docs': ['sphinx', 'sphinx_rtd_theme', 'recommonmark'],
        'tests': ['pandas']
    }
)



cpdef void state_sim(double[::1] x_state,
                     double[::1, :] llt_state,
                     const double[::1] mu_state,
                     const double[::1, :] var_state,
                     const double[::1] z_state)

cdef class KalmanTV:
    cdef int n_state, n_meas
    cdef double[::1] tmu_state
    cdef double[::1] tmu_state2
    cdef double[::1, :] tvar_state
    cdef double[::1, :] tvar_state2
    cdef double[::1, :] tvar_state3
    cdef double[::1] tmu_meas
    cdef double[::1, :] tvar_meas
    cdef double[::1, :] twgt_meas
    cdef double[::1, :] twgt_meas2
    cdef double[::1, :] llt_meas
    cdef double[::1, :] llt_state

    cpdef void predict(self,
                       double[::1] mu_state_pred,
                       double[::1, :] var_state_pred,
                       const double[::1] mu_state_past,
                       const double[::1, :] var_state_past,
                       const double[::1] mu_state,
                       const double[::1, :] wgt_state,
                       const double[::1, :] var_state)
    cpdef void update(self,
                      double[::1] mu_state_filt,
                      double[::1, :] var_state_filt,
                      const double[::1] mu_state_pred,
                      const double[::1, :] var_state_pred,
                      const double[::1] x_meas,
                      const double[::1] mu_meas,
                      const double[::1, :] wgt_meas,
                      const double[::1, :] var_meas)
    cpdef void filter(self,
                      double[::1] mu_state_pred,
                      double[::1, :] var_state_pred,
                      double[::1] mu_state_filt,
                      double[::1, :] var_state_filt,
                      const double[::1] mu_state_past,
                      const double[::1, :] var_state_past,
                      const double[::1] mu_state,
                      const double[::1, :] wgt_state,
                      const double[::1, :] var_state,
                      const double[::1] x_meas,
                      const double[::1] mu_meas,
                      const double[::1, :] wgt_meas,
                      const double[::1, :] var_meas)
    cpdef void smooth_mv(self,
                         double[::1] mu_state_smooth,
                         double[::1, :] var_state_smooth,
                         const double[::1] mu_state_next,
                         const double[::1, :] var_state_next,
                         const double[::1] mu_state_filt,
                         const double[::1, :] var_state_filt,
                         const double[::1] mu_state_pred,
                         const double[::1, :] var_state_pred,
                         const double[::1, :] wgt_state)
    cpdef void smooth_sim(self,
                          double[::1] x_state_smooth,
                          const double[::1] x_state_next,
                          const double[::1] mu_state_filt,
                          const double[::1, :] var_state_filt,
                          const double[::1] mu_state_pred,
                          const double[::1, :] var_state_pred,
                          const double[::1, :] wgt_state,
                          const double[::1] z_state)
    cpdef void smooth(self,
                      double[::1] x_state_smooth,
                      double[::1] mu_state_smooth,
                      double[::1, :] var_state_smooth,
                      const double[::1] x_state_next,
                      const double[::1] mu_state_next,
                      const double[::1, :] var_state_next,
                      const double[::1] mu_state_filt,
                      const double[::1, :] var_state_filt,
                      const double[::1] mu_state_pred,
                      const double[::1, :] var_state_pred,
                      const double[::1, :] wgt_state,
                      const double[::1] z_state)
    cpdef void forecast(self,
                        double[::1] mu_fore,
                        double[::1, :] var_fore,
                        const double[::1] mu_state_pred,
                        const double[::1, :] var_state_pred,
                        const double[::1] mu_meas,
                        const double[::1, :] wgt_meas,
                        const double[::1, :] var_meas)
            


# some useful scipy.linalg functions with numba overloading

import numpy as np
from numba import njit, types
from numba.extending import overload, register_jitable, \
    get_cython_function_address
from numba.core.errors import TypingError
from numpy.linalg import LinAlgError

# need this to overload properly
import scipy.linalg

# minimal wrapper to dpotrf
# from numba.extending import get_cython_function_address
# from numba import vectorize, njit
import ctypes
# import numpy as np
# from scipy.linalg.cython_lapack cimport dpotrf, dpotrs

# --- wrappers to BLAS/LAPACK functions ----------------------------------------

# necessary ctypes
_PTR = ctypes.POINTER
_float = ctypes.c_float
_double = ctypes.c_double
_int = ctypes.c_int
_float_p = _PTR(_float)
_double_p = _PTR(_double)
_int_p = _PTR(_int)

# dpotrf (cho_factor for doubles)
addr = get_cython_function_address('scipy.linalg.cython_lapack', 'dpotrf')
functype = ctypes.CFUNCTYPE(None,
                            _int_p,  # UPLO
                            _int_p,  # N
                            _double_p,  # U
                            _int_p,  # LDA
                            _int_p,  # INFO
                            )
_dpotrf = functype(addr)
# cho_factor: float
addr = get_cython_function_address('scipy.linalg.cython_lapack', 'spotrf')
functype = ctypes.CFUNCTYPE(None,
                            _int_p,  # UPLO
                            _int_p,  # N
                            _float_p,  # U
                            _int_p,  # LDA
                            _int_p,  # INFO
                            )
_spotrf = functype(addr)

# cho_solve: double
addr = get_cython_function_address('scipy.linalg.cython_lapack', 'dpotrs')
functype = ctypes.CFUNCTYPE(None,
                            _int_p,  # UPLO
                            _int_p,  # N
                            _int_p,  # NRHS
                            _double_p,  # U
                            _int_p,  # LDA
                            _double_p,  # X
                            _int_p,  # LDB
                            _int_p,  # INFO
                            )
_dpotrs = functype(addr)
# cho_solve: float
addr = get_cython_function_address('scipy.linalg.cython_lapack', 'spotrs')
functype = ctypes.CFUNCTYPE(None,
                            _int_p,  # UPLO
                            _int_p,  # N
                            _int_p,  # NRHS
                            _float_p,  # U
                            _int_p,  # LDA
                            _float_p,  # X
                            _int_p,  # LDB
                            _int_p,  # INFO
                            )
_spotrs = functype(addr)

# triangular matrix multiplication
# matrix-vector: double
addr = get_cython_function_address('scipy.linalg.cython_blas', 'dtrmv')
functype = ctypes.CFUNCTYPE(None,
                            _int_p,  # UPLO
                            _int_p,  # TRANS
                            _int_p,  # DIAG
                            _int_p,  # N
                            _double_p,  # A
                            _int_p,  # LDA
                            _double_p,  # X
                            _int_p,  # INCX
                            )
_dtrmv = functype(addr)
# matrix-vector: float
addr = get_cython_function_address('scipy.linalg.cython_blas', 'strmv')
functype = ctypes.CFUNCTYPE(None,
                            _int_p,  # UPLO
                            _int_p,  # TRANS
                            _int_p,  # DIAG
                            _int_p,  # N
                            _float_p,  # A
                            _int_p,  # LDA
                            _float_p,  # X
                            _int_p,  # INCX
                            )
_strmv = functype(addr)
# matrix-array: double
addr = get_cython_function_address('scipy.linalg.cython_blas', 'dtrmm')
functype = ctypes.CFUNCTYPE(None,
                            _int_p,  # SIDE
                            _int_p,  # UPLO
                            _int_p,  # TRANSA
                            _int_p,  # DIAG
                            _int_p,  # M
                            _int_p,  # N
                            _double_p,  # ALPHA
                            _double_p,  # A
                            _int_p,  # LDA
                            _double_p,  # B
                            _int_p,  # LDB
                            )
_dtrmm = functype(addr)
# matrix-array: float
addr = get_cython_function_address('scipy.linalg.cython_blas', 'strmm')
functype = ctypes.CFUNCTYPE(None,
                            _int_p,  # SIDE
                            _int_p,  # UPLO
                            _int_p,  # TRANSA
                            _int_p,  # DIAG
                            _int_p,  # M
                            _int_p,  # N
                            _float_p,  # ALPHA
                            _float_p,  # A
                            _int_p,  # LDA
                            _float_p,  # B
                            _int_p,  # LDB
                            )
_strmm = functype(addr)

# --- helper functions ---------------------------------------------------------


@register_jitable
def _check_finite_array(a):
    """
    check whether array is finite
    (copied from numba.linalg._check_finite_matrix)
    """
    for v in np.nditer(a):
        # for v in np.ravel(a):
        if not np.isfinite(v.item()):
            raise LinAlgError("Array must not contain infs or NaNs.")


@register_jitable
def _asfortranarray(a, overwrite_a=False):
    """
    Convert an array to fortran order, creating a copy if necessary and/or requested.
    That is, the function returns a reference to `a` if (1) `a` is in fortran order and (2) `overwrite_a = True`.  Otherwise, returns a reference to a copy of `a`, but in fortran-order.

    Note: `overwrite_a = False` seems to set `a1.flags.owndata = False` in njit but `True` in regular Python.
    """
    if not a.flags.f_contiguous:
        # copy unless a is f_contiguous
        a1 = np.asfortranarray(a)
    else:
        # np.asfortranarray creates a copy when c_contiguous is True
        a1 = a
    if (not overwrite_a) and a.flags.f_contiguous:
        # copy if a is f_contiguous and overwrite required
        a1 = np.copy(a)
    return a1

# --- njitable cho_factor -------------------------------------------------------


@overload(scipy.linalg.cho_factor)
def jit_cho_factor(a, lower=False, overwrite_a=False, check_finite=True):
    r"""
    njit implementation of `scipy.linalg.cho_factor()`.

    - As with the original scipy implementation, overwrite_a only "works" if the input array is in contiguous fortran order.  Otherwise a copy is created.

    Args:
        a (ndarray(dim1, dim2)): 2-dimensional matrix.
    
    Returns:
        (tuple): Cholesky factorization of matrix a and indicator if it is lower triangular.

    """

    # Reject non-ndarray types
    if not isinstance(a, types.Array):
        raise TypingError("Input must be a NumPy array.")
    # Reject ndarrays with non floating-point dtype
    if not isinstance(a.dtype, types.Float):
        raise TypingError("Input array must be of type float.")
    # Dimension check
    if a.ndim != 2:
        raise TypingError(
            "scipy.linalg.cho_factor() only supported on 2-D arrays.")
    # # Squareness check
    # if a.shape[0] != a.shape[1]:
    #     raise ValueError('Input array is expected to be square but has '
    #                      'the shape: {}.'.format(a.shape))

    # transpose constants
    UP = np.array([ord('U')], dtype=np.int32)
    LO = np.array([ord('L')], dtype=np.int32)

    # which LAPACK function to use
    _potrf = _dpotrf if a.dtype == types.float64 else _spotrf

    def cho_factor_imp(a, lower=False, overwrite_a=False, check_finite=True):
        # Reject ndarrays with unsupported dimensionality
        if a.shape[0] != a.shape[1]:
            raise LinAlgError('Input array needs to be a square matrix.')
        # Quick return for square empty array
        if a.size == 0:
            return a.copy(), lower
        # Check that matrix is finite
        if check_finite:
            _check_finite_array(a)
        # this should be a no-copy if array is already in fortran order
        # and overwrite_a = True
        a1 = _asfortranarray(a, overwrite_a)
        # determine order
        uplo = LO if lower else UP
        # wrap everything into something with ctypes
        n = np.array(a.shape[0], dtype=np.int32)
        lda = n
        info = np.empty(1, dtype=np.int32)
        # uplo = np.array([uplo], dtype=np.int32)
        # call fortran via cython_lapack
        _potrf(uplo.ctypes, n.ctypes, a1.ctypes,
               lda.ctypes, info.ctypes)
        # check return
        if info[0] > 0:
            # raise np.linalg.LinAlgError(
            #     "%d-th leading minor of the array is not positive "
            #     "definite." % info[0])
            raise LinAlgError("Input matrix is not positive definite.")
        if info[0] < 0:
            # info[0] = -info[0]
            # raise ValueError(
            #     'LAPACK reported an illegal value in %d-th argument'
            #     'on entry to "POTRF".' % info[0])
            raise ValueError(
                'LAPACK reported an illegal value on entry to "POTRF".'
            )
        return a1, lower

    return cho_factor_imp

# --- njitable cho_solve -------------------------------------------------------


@overload(scipy.linalg.cho_solve)
def jit_cho_solve(c_and_lower, b, overwrite_b=False, check_finite=True):
    r"""
    njit implementation of `scipy.linalg.cho_solve()`.

    Args:
        c_and_lower (tuple): 2-d Matrix and indicator if it is lower triangular.
        b (ndarray(dim1, dim2)): 2-d Matrix.
    
    Returns:
        (ndarray(dim1, dim2)): Solved matrix.
    """
    (c, lower) = c_and_lower
    # Reject non-ndarray types
    if not isinstance(c, types.Array) or not isinstance(b, types.Array):
        raise TypingError("c and b must be a NumPy arrays.")
    # Reject ndarrays with non floating-point dtype
    if not isinstance(c.dtype, types.Float) or not isinstance(b.dtype, types.Float):
        raise TypingError("c and b must be of type float.")
    # Dimension check
    if (c.ndim != 2) or (b.ndim > 2) or (b.ndim < 1):
        raise TypingError(
            "scipy.linalg.cho_solve() only supported on 1-D or 2-D arrays.")

    # transpose constants
    UP = np.array([ord('U')], dtype=np.int32)
    LO = np.array([ord('L')], dtype=np.int32)

    # which LAPACK function to use
    _potrs = \
        _dpotrs if (c.dtype == types.float64) or (b.dtype == types.float64) \
        else _spotrs
    # whether to promote to float64
    to_float64 = c.dtype != b.dtype

    def cho_solve_imp(c_and_lower, b, overwrite_b=False, check_finite=True):
        (c, lower) = c_and_lower
        # Squareness check
        if (c.shape[0] != c.shape[1]):
            raise LinAlgError("The factored matrix c is not square.")
        # Dimension check
        if c.shape[1] != b.shape[0]:
            raise LinAlgError("c and b have incompatible dimensions.")
        # Check that arrays are finite
        if check_finite:
            _check_finite_array(c)
            _check_finite_array(b)
        if to_float64:
            # need to upcast these. if already float64 should be a no-copy.
            # otherwise, copies to C order, so do this before as_fortranarray
            c1 = np.asarray(c, dtype=np.float64)
            b1 = np.asarray(b, dtype=np.float64)
        else:
            c1 = c
            b1 = b
        # convert to fortran order, avoiding copy whenever possible.
        # c is assumed to come from cho_factor, so already in fortran order.
        # if it was already float64, _asfortranarray does nothing
        # if it needed to be changed to float64, a copy was already made,
        # so _asfortranarray is not overwriting the original c.
        c1 = _asfortranarray(c1, overwrite_a=True)
        b1 = _asfortranarray(b1, overwrite_b)
        # determine order
        uplo = LO if lower else UP
        # wrap everything into something with ctypes
        n = np.array(c.shape[0], dtype=np.int32)
        # nrhs = np.array(b.shape[1], dtype=np.int32)
        nrhs = np.array(b.size/b.shape[0], dtype=np.int32)
        lda = n
        ldb = n
        info = np.empty(1, dtype=np.int32)
        # call fortran via cython_lapack
        _potrs(uplo.ctypes, n.ctypes, nrhs.ctypes, c1.ctypes, lda.ctypes,
               b1.ctypes, ldb.ctypes, info.ctypes)
        # check return
        if info[0] != 0:
            raise ValueError(
                'Illegal value in internal "POTRS".'
            )
        return b1

    return cho_solve_imp

# --- njitable triangular matrix multiplication (tri_mult) ---------------------


def tri_mult(a_and_lower, x, overwrite_x=False, check_finite=True):
    r"""
    Triangular matrix multiplication :math:`a x`.

    This provides an unoptimized Python implementation for the sole purpose of testing unjitted code.  In the pure Python version, both `overwrite_x` and `check_finite` flags are ignored.
    """
    (a, lower) = a_and_lower
    # start by converting to fortran order
    a1 = np.asfortranarray(a)
    x1 = np.asfortranarray(x)
    # since a is typically coming from cho_factor
    # might have trash in the part that's not referenced.
    if lower:
        a1 = scipy.linalg.tril(a1)
    else:
        a1 = scipy.linalg.triu(a1)
    # reshape x into a matrix
    x1 = x1.reshape((x1.shape[0], -1), order="F")
    if a1.dtype != x1.dtype:
        # convert both to float64
        a1 = np.asarray(a1, dtype=np.float64)
        x1 = np.asarray(x1, dtype=np.float64)
    y = np.dot(a1, x1)
    # back to original shape
    y = y.reshape(x.shape, order="F")
    if overwrite_x:
        # inefficient, but gives the correct result
        x[:] = y
        y = x
    return y


@overload(tri_mult)
def jit_tri_mult(a_and_lower, x, overwrite_x=False, check_finite=True):
    r"""
    Triangular matrix multiplication :math:`a x`.
    
    Args:
        a_and_lower (tuple):  2-d Matrix and indicator if it is lower triangular.
        x (ndarray(dim1, dim2)): 2-d Matrix.
    
    Returns:
        (ndarray(dim1, dim2)): :math:`a x`.
    """
    (a, lower) = a_and_lower
    # Reject non-ndarray types
    if not isinstance(a, types.Array) or not isinstance(x, types.Array):
        raise TypingError("a and x must be a NumPy arrays.")
    # Reject ndarrays with non floating-point dtype
    if not isinstance(a.dtype, types.Float) or not isinstance(x.dtype, types.Float):
        raise TypingError("a and x must be of type float.")

    # type to use
    use_float64 = (a.dtype == types.float64) or (x.dtype == types.float64)
    # whether to promote to float64
    to_float64 = a.dtype != x.dtype
    # print("use_float64 = ", use_float64)
    # print("to_float64 = ", to_float64)

    # BLAS constants
    UP = np.array([ord('U')], dtype=np.int32)
    LO = np.array([ord('L')], dtype=np.int32)
    trans = np.array([ord('N')], dtype=np.int32)
    diag = np.array([ord('N')], dtype=np.int32)
    incx = np.array([1], dtype=np.int32)
    side = np.array([ord('L')], dtype=np.int32)
    alpha = np.array(1.0, dtype=np.float64 if use_float64 else np.float32)

    # Which BLAS function to use
    if x.ndim == 1:
        _trmv = _dtrmv if use_float64 else _strmv
    else:
        _trmm = _dtrmm if use_float64 else _strmm

    # implementation for matrix-vector multiplication
    def tri_mult_imp_1D(a_and_lower, x, overwrite_x=False, check_finite=True):
        (a, lower) = a_and_lower
        # Squareness check
        if (a.ndim != 2) or (a.shape[0] != a.shape[1]):
            raise LinAlgError("The triangular matrix a is not square.")
        # Dimension check
        if a.shape[1] != x.shape[0]:
            raise LinAlgError("a and x have incompatible dimensions.")
        # Check that arrays are finite
        if check_finite:
            _check_finite_array(a)
            _check_finite_array(x)
        # float and fortran conversions
        if to_float64:
            # need to upcast these. if already float64 should be a no-copy.
            # otherwise, copies to C order, so do this before as_fortranarray
            a1 = np.asarray(a, dtype=np.float64)
            x1 = np.asarray(x, dtype=np.float64)
        else:
            a1 = a
            x1 = x
        # convert to fortran order, avoiding copy whenever possible.
        a1 = _asfortranarray(a1, overwrite_a=True)
        x1 = _asfortranarray(x1, overwrite_x)
        # determine order
        uplo = LO if lower else UP
        # wrap everything into something with ctypes
        n = np.array(a.shape[0], dtype=np.int32)
        lda = n
        # call fortran via cython_blas
        _trmv(uplo.ctypes, trans.ctypes, diag.ctypes, n.ctypes,
              a1.ctypes, lda.ctypes, x1.ctypes, incx.ctypes)
        return x1

    # implementation for matrix-(array >1D) multiplication
    def tri_mult_imp_ND(a_and_lower, x, overwrite_x=False, check_finite=True):
        (a, lower) = a_and_lower
        # Squareness check
        if (a.ndim != 2) or (a.shape[0] != a.shape[1]):
            raise LinAlgError("The triangular matrix a is not square.")
        # Dimension check
        if a.shape[1] != x.shape[0]:
            raise LinAlgError("a and x have incompatible dimensions.")
        # Check that arrays are finite
        if check_finite:
            _check_finite_array(a)
            _check_finite_array(x)
        # float and fortran conversions
        if to_float64:
            # need to upcast these. if already float64 should be a no-copy.
            # otherwise, copies to C order, so do this before as_fortranarray
            a1 = np.asarray(a, dtype=np.float64)
            x1 = np.asarray(x, dtype=np.float64)
        else:
            a1 = a
            x1 = x
        # convert to fortran order, avoiding copy whenever possible.
        a1 = _asfortranarray(a1, overwrite_a=True)
        x1 = _asfortranarray(x1, overwrite_x)
        # determine order
        uplo = LO if lower else UP
        # wrap everything into something with ctypes
        m = np.array(a.shape[0], dtype=np.int32)
        n = np.array(x.size/x.shape[0], dtype=np.int32)
        lda = m
        ldb = m
        # call fortran via cython_blas
        _trmm(side.ctypes, uplo.ctypes, trans.ctypes, diag.ctypes,
              m.ctypes, n.ctypes, alpha.ctypes,
              a1.ctypes, lda.ctypes, x1.ctypes, ldb.ctypes)
        return x1

    if x.ndim == 1:
        return tri_mult_imp_1D
    else:
        return tri_mult_imp_ND



cdef extern from "KalmanTV_raw.h" namespace "kalmantv_raw":
    cdef cppclass KalmanTV_raw:
        KalmanTV_raw(int, int) except +
        void predict(double * mu_state_pred,
                     double * var_state_pred,
                     const double * mu_state_past,
                     const double * var_state_past,
                     const double * mu_state,
                     const double * wgt_state,
                     const double * var_state)
        void update(double * mu_state_filt,
                    double * var_state_filt,
                    const double * mu_state_pred,
                    const double * var_state_pred,
                    const double * x_meas,
                    const double * mu_meas,
                    const double * wgt_meas,
                    const double * var_meas)
        void filter(double * mu_state_pred,
                    double * var_state_pred,
                    double * mu_state_filt,
                    double * var_state_filt,
                    const double * mu_state_past,
                    const double * var_state_past,
                    const double * mu_state,
                    const double * wgt_state,
                    const double * var_state,
                    const double * x_meas,
                    const double * mu_meas,
                    const double * wgt_meas,
                    const double * var_meas)
        void smooth_mv(double * mu_state_smooth,
                       double * var_state_smooth,
                       const double * mu_state_next,
                       const double * var_state_next,
                       const double * mu_state_filt,
                       const double * var_state_filt,
                       const double * mu_state_pred,
                       const double * var_state_pred,
                       const double * wgt_state)
        void smooth_sim(double * x_state_smooth,
                        const double * x_state_next,
                        const double * mu_state_filt,
                        const double * var_state_filt,
                        const double * mu_state_pred,
                        const double * var_state_pred,
                        const double * wgt_state,
                        const double * z_state)
        void smooth(double * x_state_smooth,
                    double * mu_state_smooth,
                    double * var_state_smooth,
                    const double * x_state_next,
                    const double * mu_state_next,
                    const double * var_state_next,
                    const double * mu_state_filt,
                    const double * var_state_filt,
                    const double * mu_state_pred,
                    const double * var_state_pred,
                    const double * wgt_state,
                    const double * z_state)
        void state_sim(double * x_state,
                       const double * mu_state,
                       const double * var_state,
                       const double * z_state)
        void forecast(double * mu_fore,
                      double * var_fore,
                      const double * mu_state_pred,
                      const double * var_state_pred,
                      const double * mu_meas,
                      const double * wgt_meas,
                      const double * var_meas)



import numpy as np
import scipy.linalg

def _solveV(V, B=None):
    """
    Computes :math:`X = V^{-1}B` where V is a variance matrix.

    Args:
        V (ndarray(n_dim1, n_dim1)): Variance matrix V in :math:`X = V^{-1}B`.
        B (ndarray(n_dim1, n_dim2)): Optional Matrix B in :math:`X = V^{-1}B`.  If `None` defaults to the `n_dim1 x n_dim1` identity matrix.

    Returns:
        (ndarray(n_dim1, n_dim2)): Matrix X in :math:`X = V^{-1}B`.

    """
    if B is None:
        B = np.identity(V.shape[0])
    L, low = scipy.linalg.cho_factor(V)
    return scipy.linalg.cho_solve((L, low), B)


def mvncond(mu, Sigma, icond):
    """
    Calculates A, b, and V such that :math:`y[!icond] | y[icond] \sim N(A y[icond] + b, V)`.

    Args:
        mu (ndarray(2*n_dim)): Mean of y.
        Sigma (ndarray(2*n_dim, 2*n_dim)): Covariance of y. 
        icond (ndarray(2*nd_dim)): Conditioning on the terms given.

    Returns:
        (tuple):
        - **A** (ndarray(n_dim, n_dim)): For :math:`y \sim N(\mu, \Sigma)` 
          such that :math:`y[!icond] | y[icond] \sim N(A y[icond] + b, V)` Calculate A.
        - **b** (ndarray(n_dim)): For :math:`y \sim N(\mu, \Sigma)` 
          such that :math:`y[!icond] | y[icond] \sim N(A y[icond] + b, V)` Calculate b.
        - **V** (ndarray(n_dim, n_dim)): For :math:`y \sim N(\mu, \Sigma)`
          such that :math:`y[!icond] | y[icond] \sim N(A y[icond] + b, V)` Calculate V.

    """
    # if y1 = y[~icond] and y2 = y[icond], should have A = Sigma12 * Sigma22^{-1}
    # A = np.dot(Sigma[np.ix_(~icond, icond)],
    #            sp.linalg.cho_solve(
    #                sp.linalg.cho_factor(Sigma[np.ix_(icond, icond)]),
    #                np.identity(sum(icond)))
    #            )
    Sigma12 = Sigma[np.ix_(~icond, icond)]
    A = np.dot(Sigma12, _solveV(Sigma[np.ix_(icond, icond)]))
    b = mu[~icond] - np.dot(A, mu[icond])  # mu1 - A * mu2
    V = Sigma[np.ix_(~icond, ~icond)] - \
        np.dot(A, Sigma12.T)  # Sigma11 - A * Sigma21
    return A, b, V



cpdef void vec_copy(double[::1] y,
                    const double[::1] x)
cpdef void mat_copy(double[::1, :] B,
                    const double[::1, :] A)
cpdef void vec_add(double[::1] y,
                   const double alpha, 
                   const double[::1] x)
cpdef void mat_add(double[::1, :] B,
                   const double alpha,
                   const double beta,
                   const double[::1, :] A)
cpdef void mat_vec_mult(double[::1] y,
                        char* trans,
                        const double alpha, 
                        const double beta,
                        const double[::1, :] A, 
                        const double[::1] x)
cpdef void tri_vec_mult(const double[::1] x, 
                        char* uplo,
                        char* trans,
                        char* diag, 
                        const double[::1, :] A)
cpdef void mat_mult(double[::1, :] C,
                    char* transa, 
                    char* transb, 
                    const double alpha, 
                    const double beta,
                    const double[::1, :] A, 
                    const double[::1, :] B)
cpdef void mat_triple_mult(double[::1, :] D,
                           double[::1, :] temp, 
                           char* transa, 
                           char* transb, 
                           char* transc, 
                           const double alpha, 
                           const double beta,
                           const double[::1, :] A, 
                           const double[::1, :] B, 
                           const double[::1, :] C)
cpdef void chol_fact(double[::1, :] U,
                     const double[::1, :] V)
cpdef void solveV(double[::1, :] U,
                  double[::1, :] X,
                  const double[::1, :] V, 
                  const double[::1, :] B)
                  